import os
import json
import csv
import shutil
import pickle
import datetime
import pkg_resources
import numpy as np

from PySide6.QtWidgets import QMainWindow, QToolBar, QDockWidget, \
    QMessageBox, QFileDialog, QLabel, QMenu
from PySide6.QtCore import Qt, Slot, QUrl, QDir, Signal, QObject, QThread
from PySide6.QtWebChannel import QWebChannel
from PySide6.QtGui import QIcon, QPixmap

from ..utils.common import get_immediate_subdirectories

from ..ui.ui_mainwindow import Ui_MainWindow
from .map import MapView, ColorbarView, DataColumnSelectionView, \
    DataRangeView, ColormapSelectionView, LayerSelectionView
from .annotation_editor import AnnotationEditorView
from .data_sources import DataSourcesView
from .source_frame import SourceFrameView
from .patches import PatchesView
from .analysis import AnalysisView
from .analysis_details import AnalysisDetailsView
from .string_editor import StringEditorView
from .dataset_settings import DatasetSettingsView


class MainView(QMainWindow):
    def __init__(self, model, controller):
        super().__init__()
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        self.model = model
        self.controller = controller

        # set window icon
        app_icon = QIcon(QPixmap(pkg_resources.resource_filename("src.resources", "app_icon.png")))
        self.setWindowIcon(app_icon)

        # add icons
        self.ui.actionOpen_Dataset.setIcon(QIcon.fromTheme("document-open"))
        self.ui.actionClose_Dataset.setIcon(QIcon.fromTheme("window-close"))
        self.ui.actionDataset_Settings.setIcon(QIcon.fromTheme("document-properties"))        
        self.ui.actionQuit.setIcon(QIcon.fromTheme("application-exit"))
        self.ui.actionNew_Defect_Annotation.setIcon(QIcon.fromTheme("document-new"))
        self.ui.actionLoad_Defect_Annotation.setIcon(QIcon.fromTheme("document-open"))
        self.ui.actionSave_Defect_Annotation.setIcon(QIcon.fromTheme("document-save"))
        self.ui.actionClose_Defect_Annotation.setIcon(QIcon.fromTheme("window-close"))
        self.ui.actionExport_String_Annotation.setIcon(QIcon.fromTheme("document-save"))
        self.ui.actionClose_String_Annotation.setIcon(QIcon.fromTheme("window-close"))
        self.ui.actionAbout.setIcon(QIcon.fromTheme("help-about"))        
        
        # register map view
        self.map_view = MapView(model, controller, parent=self)
        self.channel = QWebChannel(self.ui.widget.page())
        self.ui.widget.page().setWebChannel(self.channel)
        self.ui.widget.setContextMenuPolicy(Qt.NoContextMenu)
        self.channel.registerObject("map_view", self.map_view)

        # register string editor controller with web engine
        self.channel.registerObject("string_editor_controller", self.controller.string_editor_controller)

        # add colorbar for map view
        self.colorbarView = ColorbarView(self.model, self.controller)
        self.ui.gridLayout.addWidget(self.colorbarView.widget, 1, 0, 1, 1)

        # setup toolbars
        self.toolBarDataColumnSelection = QToolBar(u"Data Selection Toolbar", self)
        self.addToolBar(Qt.TopToolBarArea, self.toolBarDataColumnSelection)
        self.dataColumnSelectionView = DataColumnSelectionView(self.model, self.controller, parent=self)
        self.toolBarDataColumnSelection.addWidget(self.dataColumnSelectionView)

        self.toolBarDataRange = QToolBar(u"Data Range Toolbar", self)
        self.addToolBar(Qt.TopToolBarArea, self.toolBarDataRange)
        self.dataRangeWidget = DataRangeView(self.model, self.controller, parent=self)
        self.toolBarDataRange.addWidget(self.dataRangeWidget)

        self.toolBarColormapSelection = QToolBar(u"Colormap Selection Toolbar", self)
        self.addToolBar(Qt.TopToolBarArea, self.toolBarColormapSelection)
        self.colormapSelectionWidget = ColormapSelectionView(self.model, self.controller, parent=self)
        self.toolBarColormapSelection.addWidget(self.colormapSelectionWidget)

        self.toolBarLayerSelection = QToolBar(u"Layer Selection Toolbar", self)
        self.addToolBar(Qt.TopToolBarArea, self.toolBarLayerSelection)
        self.layerSelectionWidget = LayerSelectionView(self.model, self.controller, parent=self)
        self.toolBarLayerSelection.addWidget(self.layerSelectionWidget)

        # setup widgets
        self.annotationEditorWidget = QDockWidget(u"Annotation Editor", self)
        self.annotation_editor = AnnotationEditorView(self.model, self.controller, parent=self)
        self.annotationEditorWidget.setWidget(self.annotation_editor)
        
        self.stringEditorWidget = QDockWidget(u"String Editor", self)
        self.string_editor = StringEditorView(self.model, self.controller, parent=self)
        self.stringEditorWidget.setWidget(self.string_editor)

        self.sourceFrameWidget = QDockWidget(u"Source Frame", self)
        self.source_frame = SourceFrameView(self.model, self.controller, parent=self)
        self.sourceFrameWidget.setWidget(self.source_frame)

        self.patchesWidget = QDockWidget(u"Patches", self)
        self.patches = PatchesView(self.model, self.controller, parent=self)
        self.patchesWidget.setWidget(self.patches)
        
        self.dataSourcesWidget = QDockWidget(u"Data Sources", self)
        self.data_sources = DataSourcesView(self.model, self.controller, parent=self)
        self.dataSourcesWidget.setWidget(self.data_sources)

        self.addDockWidget(Qt.LeftDockWidgetArea, self.stringEditorWidget)
        self.addDockWidget(Qt.LeftDockWidgetArea, self.dataSourcesWidget)
        self.addDockWidget(Qt.LeftDockWidgetArea, self.annotationEditorWidget)
        self.addDockWidget(Qt.LeftDockWidgetArea, self.sourceFrameWidget)
        self.addDockWidget(Qt.LeftDockWidgetArea, self.patchesWidget)
        self.tabifyDockWidget(self.patchesWidget, self.sourceFrameWidget)

        # setup status bar
        self.moduleIdLabel = QLabel()
        self.ui.statusBar.addPermanentWidget(self.moduleIdLabel)
        self.datasetInfoLabel = QLabel()
        self.ui.statusBar.addPermanentWidget(self.datasetInfoLabel)

        # file menu
        self.ui.actionOpen_Dataset.triggered.connect(self.open_dataset)
        self.ui.actionClose_Dataset.triggered.connect(self.close_dataset)
        self.ui.actionDataset_Settings.triggered.connect(lambda: self.show_child_window("dataset_settings"))
        self.ui.actionQuit.triggered.connect(self.close)

        # annotation menu
        self.ui.actionNew_Defect_Annotation.triggered.connect(self.new_defect_annotation)
        self.ui.actionLoad_Defect_Annotation.triggered.connect(self.load_defect_annotation)
        self.ui.actionSave_Defect_Annotation.triggered.connect(self.save_defect_annotation)
        self.ui.actionClose_Defect_Annotation.triggered.connect(self.close_defect_annotation)
        self.ui.actionAnnotate_Strings.triggered.connect(self.annotate_strings)
        self.ui.actionExport_String_Annotation.triggered.connect(self.export_string_annotation)
        self.ui.actionClose_String_Annotation.triggered.connect(self.close_string_annotation)

        # analysis menu
        self.ui.actionNew_Analysis.triggered.connect(lambda: self.show_child_window("analysis"))

        # view menu
        self.ui.menuView.addAction(self.dataSourcesWidget.toggleViewAction())
        self.ui.menuView.addAction(self.stringEditorWidget.toggleViewAction())
        self.ui.menuView.addAction(self.annotationEditorWidget.toggleViewAction())
        self.ui.menuView.addAction(self.sourceFrameWidget.toggleViewAction())
        self.ui.menuView.addAction(self.patchesWidget.toggleViewAction())
        self.toolbar_view_menu = QMenu(u"Toolbars")
        self.toolbar_view_menu.addAction(self.toolBarDataColumnSelection.toggleViewAction())
        self.toolbar_view_menu.addAction(self.toolBarDataRange.toggleViewAction())
        self.toolbar_view_menu.addAction(self.toolBarColormapSelection.toggleViewAction())
        self.toolbar_view_menu.addAction(self.toolBarLayerSelection.toggleViewAction())        
        self.ui.menuView.addMenu(self.toolbar_view_menu)

        # about menu
        self.ui.actionAbout.triggered.connect(self.about)

        # connect signals and slots
        self.model.dataset_opened.connect(self.dataset_opened)
        self.model.dataset_closed.connect(self.dataset_closed)
        self.model.app_mode_changed.connect(self.app_mode_changed)
        self.model.annotation_editor_model.has_changes_changed.connect(self.defect_annotation_has_changes)
        self.model.annotation_editor_model.current_file_name_changed.connect(self.defect_annotation_has_changes)
        self.controller.annotation_editor_controller.close_dataset.connect(self.controller.close_dataset)
        self.model.dataset_stats_changed.connect(self.update_status_bar)
        self.model.track_id_changed.connect(self.update_status_bar)
        self.model.string_editor_model.string_annotation_data_changed.connect(self.update_status_bar)
        
        # load HTML document for map view
        index_file = QDir.current().filePath(pkg_resources.resource_filename("src", "index.html"))
        index_url = QUrl.fromLocalFile(index_file)
        self.ui.widget.load(index_url)

        self.update_status_bar()
        self.child_windows = {}

        # set defaults
        self.model.app_mode = None

        # for development
        #dir = "/home/lukas/HI-ERN-2020/Dataset-Viewer-Georeferencing-Desktop/test_data"
        #self.controller.open_dataset(dir)

    def valid_dataset(self, dir):
        if not os.path.isfile(os.path.join(dir, "version.json")):
            return False
        probe_dirs = get_immediate_subdirectories(dir)
        if not "mapping" in probe_dirs:
            return False
        if not "patches" in probe_dirs:
            return False
        if not "splitted" in probe_dirs:
            return False
        return True

    @Slot()
    def open_dataset(self):
        dir = QFileDialog.getExistingDirectory(
            self, caption="Open Dataset", options=QFileDialog.ShowDirsOnly)
        if dir == "":
            return
        if not self.valid_dataset(dir):
            msg = QMessageBox()
            msg.setWindowTitle("Error")
            msg.setText("Not a valid PV Hawk Dataset.")
            msg.setIcon(QMessageBox.Critical)
            msg.exec()
            return
        self.controller.open_dataset(dir)

    def close_dataset(self):
        self.controller.stop_background_threads()
        self.controller.close_dataset_request()

    def dataset_opened(self):
        self.ui.actionClose_Dataset.setEnabled(True)
        self.ui.actionOpen_Dataset.setEnabled(False)
        self.ui.actionDataset_Settings.setEnabled(True)
        self.ui.actionNew_Analysis.setEnabled(True)
        self.ui.actionNew_Defect_Annotation.setEnabled(True)
        self.ui.actionLoad_Defect_Annotation.setEnabled(True)
        self.ui.actionAnnotate_Strings.setEnabled(True)
        self.ui.actionExport_String_Annotation.setEnabled(True)
        self.ui.actionClose_String_Annotation.setEnabled(False)
        self.ui.statusBar.showMessage("Dataset opened", 5000)

    def dataset_closed(self):
        self.ui.actionClose_Dataset.setEnabled(False)
        self.ui.actionOpen_Dataset.setEnabled(True)
        self.ui.actionDataset_Settings.setEnabled(False)
        self.ui.actionNew_Analysis.setEnabled(False)
        self.ui.actionNew_Defect_Annotation.setEnabled(False)
        self.ui.actionLoad_Defect_Annotation.setEnabled(False)
        self.ui.actionSave_Defect_Annotation.setEnabled(False)
        self.ui.actionClose_Defect_Annotation.setEnabled(False)
        self.ui.actionAnnotate_Strings.setEnabled(False)
        self.ui.actionExport_String_Annotation.setEnabled(False)
        self.ui.actionClose_String_Annotation.setEnabled(False)
        self.ui.statusBar.showMessage("Dataset closed", 5000)

    @Slot()
    def update_status_bar(self):
        dataset_dir = self.model.dataset_dir
        dataset_info = ""
        dataset_info_text = ""
        module_id_text = ""
        if dataset_dir is not None:
            
            dataset_info_text = "Dataset Info"
            stats = self.model.dataset_stats
            if stats is None:
                stats = {
                    "num_modules": 0,
                    "num_patches": 0,
                    "flight_duration": "00:00:00.000",
                    "trajectory_length": 0,
                }

            dataset_info = (
                "Dataset Path: {}<br>".format(dataset_dir) + 
                "Num. Modules: {}<br>".format(stats["num_modules"]) +
                "Num. Patches: {}<br>".format(stats["num_patches"]) + 
                "Flight duration: {}<br>".format(stats["flight_duration"]) +
                "Trajectory length: {} m".format(int(stats["trajectory_length"]))
            )
	
            # set module ID label
            track_id = self.model.track_id
            module_id_text = ""
            if track_id is not None:
                module_id_text = "{}".format(track_id)
            
                # search for plant ID
                string_annotation_data = self.model.string_editor_model.string_annotation_data
                if string_annotation_data is not None:
                    id_mapping = self.model.string_editor_model.string_annotation_data["plant_id_track_id_mapping"]
                    id_mapping = {track_id_: plant_id_ for plant_id_, track_id_ in id_mapping}
                    try:
                        plant_id = id_mapping[track_id]                
                    except KeyError:
                        pass
                    else:
                        module_id_text = "{} ({})".format(module_id_text, plant_id)
                        
        self.datasetInfoLabel.setText(dataset_info_text)
        self.datasetInfoLabel.setToolTip(dataset_info)
        self.moduleIdLabel.setText(module_id_text)

    @Slot(str)
    def app_mode_changed(self, app_mode):
        print("app_mode ", app_mode)
        if app_mode == "defect_annotation":
            self.annotationEditorWidget.show()
            self.annotationEditorWidget.setEnabled(True)
            self.ui.actionClose_Defect_Annotation.setEnabled(True)
            self.defect_annotation_has_changes(False)
        else:
            self.annotationEditorWidget.hide()
            self.annotationEditorWidget.setEnabled(False)
            self.ui.actionSave_Defect_Annotation.setEnabled(False)
            self.ui.actionClose_Defect_Annotation.setEnabled(False)

        if app_mode == "string_annotation":
            self.stringEditorWidget.show()
            self.stringEditorWidget.setEnabled(True)
            self.ui.actionClose_String_Annotation.setEnabled(True)
        else:
            self.stringEditorWidget.hide()
            self.stringEditorWidget.setEnabled(False)
            self.ui.actionClose_String_Annotation.setEnabled(False)

        if app_mode is None or app_mode == "data_visualization":
            self.setWindowTitle("PV Hawk Viewer")
        

    @Slot(bool)
    def defect_annotation_has_changes(self, has_changes):
        if self.model.app_mode != "defect_annotation":
            return
        file_name = os.path.basename(self.model.annotation_editor_model.current_file_name)
        if has_changes:
            self.ui.actionSave_Defect_Annotation.setEnabled(True)            
            self.setWindowTitle("PV Hawk Viewer - {}*".format(file_name))
        else:
            self.ui.actionSave_Defect_Annotation.setEnabled(False)
            self.setWindowTitle("PV Hawk Viewer - {}".format(file_name))

    @Slot()
    def new_defect_annotation(self):
        self.model.app_mode = "defect_annotation"
        self.controller.new_defect_annotation.emit()

    @Slot()
    def load_defect_annotation(self):
        self.controller.load_defect_annotation.emit()

    @Slot()
    def save_defect_annotation(self):
        self.controller.save_defect_annotation.emit()

    @Slot()
    def close_defect_annotation(self):
        self.controller.close_defect_annotation.emit()

    @Slot()
    def annotate_strings(self):
        if self.model.app_mode == "defect_annotation":
            self.controller.close_defect_annotation.emit()
        self.model.app_mode = "string_annotation"
        # remark on line above: in close defect annotation app mode is set to
        # "data_visualization". Nothing guarantees that the subsequent line
        # is executed after the app mode was set to "data_visualization".

    @Slot()
    def export_string_annotation(self):
        self.controller.export_string_annotation.emit()

    @Slot()
    def close_string_annotation(self):
        self.controller.close_string_annotation.emit()
        self.model.app_mode = "data_visualization"

    @Slot()
    def show_child_window(self, which):
        if which == "analysis":
            if not self.model.dataset_is_open:
                return
            if which not in self.child_windows:
                self.child_windows[which] = AnalysisView(self.model, self.controller, self)
            self.controller.analysis_controller.reset()
            self.child_windows[which].show()

        elif which == "dataset_settings":
            if not self.model.dataset_is_open:
                return
            if which not in self.child_windows:
                self.child_windows[which] = DatasetSettingsView(self.model, self.controller, self)
            self.controller.analysis_controller.reset()
            self.child_windows[which].show()

        elif which == "analysis_details":
            if self.model.meta is None:
                return
            if which not in self.child_windows:
                self.child_windows[which] = AnalysisDetailsView(self.model, self.controller, self)
            self.child_windows[which].show()

    def about(self):
        gh1 = "LukasBommes/PV-Hawk"
        gh2 = "LukasBommes/PV-Hawk-Viewer"
        about_text = "PV Hawk Viewer<br><br>" \
            + "Author: Lukas Bommes<br>" \
            + "Organization: Helmholtz Institute Erlangen-Nürnberg for Renewable Energy (HI ERN)<br>" \
            + "GitHub:<br>" \
            + "PV Hawk: <a href='https://github.com/{gh1}'>{gh1}</a><br>".format(gh1=gh1) \
            + "PV Hawk Viewer: <a href='https://github.com/{gh2}'>{gh2}</a><br>".format(gh2=gh2)
        QMessageBox.about(
            self,
            "About PV Hawk Viewer",
            about_text
        )

    def closeEvent(self, event):
        """Ask whether unsaved changes should be saved"""
        self.controller.stop_background_threads()
        self.controller.mainwindow_close_requested.emit(event)        



class MainController(QObject):
    source_deleted = Signal()
    new_defect_annotation = Signal()
    save_defect_annotation = Signal()
    load_defect_annotation = Signal()
    close_defect_annotation = Signal()
    export_string_annotation = Signal()
    close_string_annotation = Signal()
    mainwindow_close_requested = Signal(object)
    dataset_close_requested = Signal()
    redraw_map = Signal()

    def __init__(self, model):
        super().__init__()
        self.model = model
        self.thread_dataset_stats = None
        self.worker_dataset_stats = None

    def reset(self):
        self.model.dataset_dir = None
        self.model.dataset_version = None
        self.model.data = None
        self.model.meta = None
        self.model.patch_meta = None
        self.model.sun_reflections = None
        self.model.track_ids = None
        self.model.app_mode = None
        self.model.source_names = None
        self.model.dataset_is_open = False
        self.model.selected_source = None
        self.model.selected_column = None
        self.model.track_id = None
        self.model.dataset_stats = None

    @Slot(str)
    def open_dataset(self, dataset_dir):
        self.model.dataset_dir = dataset_dir
        version_info = json.load(open(os.path.join(self.model.dataset_dir, "version.json"), "r"))
        self.model.dataset_version = version_info["dataset_version"]
        if self.model.dataset_version == "v1":
            self.model.patch_meta = pickle.load(open(os.path.join(
                self.model.dataset_dir, "patches", "meta.pkl"), "rb"))
        elif self.model.dataset_version == "v2":
            self.model.patch_meta = pickle.load(open(os.path.join(
                self.model.dataset_dir, "quadrilaterals", "quadrilaterals.pkl"), "rb"))
        self.load_dataset_settings()
        self.load_sun_reflections()
        self.update_source_names()
        self.load_source("Module Layout")
        self.update_track_ids()
        self.update_dataset_stats()
        self.model.dataset_is_open = True
        self.model.app_mode = "data_visualization"

    @Slot()
    def close_dataset_request(self):
        if self.model.app_mode == "defect_annotation":
            self.dataset_close_requested.emit()
        #elif self.model.app_mode == "string_annotation":
        #    pass
        else:
            self.close_dataset()

    @Slot()
    def close_dataset(self):
        self.reset()

    @Slot()
    def update_source_names(self):
        source_names = []
        if self.model.dataset_dir is not None:
            try:
                source_names = sorted(get_immediate_subdirectories(
                    os.path.join(self.model.dataset_dir, "analyses")))
            except FileNotFoundError:
                pass
        source_names.insert(0, "Module Layout")
        try:
            source_names.remove("Sun Filter")  # ignore sun_filter output in analysis
        except ValueError:
            pass
        self.model.source_names = source_names

    @Slot()
    def update_track_ids(self):
        self.model.track_ids = list(self.get_column("track_id").values())

    def load_sun_reflections(self):
        if self.model.dataset_dir is None:
            return
        try:
            self.model.sun_reflections = json.load(open(os.path.join(
                self.model.dataset_dir, "analyses", "Sun Filter", "sun_filter.json"), "r"))
        except FileNotFoundError:
            pass
    
    @Slot()
    def load_dataset_settings(self):
        if self.model.dataset_dir is None:
            return
        try:
            settings = json.load(open(os.path.join(self.model.dataset_dir, "settings.json"), "r"))
        except FileNotFoundError:
            # create settings file with defaults
            self.model.dataset_settings_model.gain = 0.04
            self.model.dataset_settings_model.offset = -273.15
            self.save_dataset_settings()
        else:
            self.model.dataset_settings_model.gain = settings["raw_image_to_celsius"]["gain"]
            self.model.dataset_settings_model.offset = settings["raw_image_to_celsius"]["offset"]

    @Slot()
    def save_dataset_settings(self):
        if self.model.dataset_dir is None:
            return
        settings = {
            "raw_image_to_celsius": {
                "gain": self.model.dataset_settings_model.gain,
                "offset": self.model.dataset_settings_model.offset
            }
        }
        json.dump(settings, open(os.path.join(self.model.dataset_dir, "settings.json"), "w"))

    @Slot(str)
    def load_source(self, selected_source):
        print("Updating", selected_source)
        if self.model.dataset_dir is None:
            return
        if selected_source is None:
            return
        if selected_source == "Module Layout":
            self.model.data = json.load(open(os.path.join(
                self.model.dataset_dir, "mapping", "module_geolocations_refined.geojson"), "r"))
            self.model.meta = None            
        else:
            self.model.data = json.load(open(os.path.join(
                self.model.dataset_dir, "analyses", selected_source, "results.geojson"), "r"))
            self.model.meta = json.load(open(os.path.join(
                self.model.dataset_dir, "analyses", selected_source, "meta.json"), "r"))
        self.model.selected_source = selected_source

    @Slot()
    def delete_source(self, selected_source):
        if self.model.dataset_dir is None:
            return
        if selected_source is None:
            return
        if selected_source == "Module Layout":
            return
        if selected_source == self.model.selected_source:
            self.source_deleted.emit()
        rmdir = os.path.join(self.model.dataset_dir, "analyses", selected_source)
        print("Deleting {}".format(rmdir))
        shutil.rmtree(rmdir, ignore_errors=True)
        self.update_source_names()

    @Slot(int)
    def set_selected_column(self, value):
        # first update map value range then set selected column
        columns_names = self.get_column_names()
        try:
            column = columns_names[value]
        except IndexError:
            min_val = -5
            max_val = 5
        else:
            data_column = self.get_column(column)
            if len(data_column) == 0:
                min_val = -5
                max_val = 5
            else:
                min_val = min(data_column.values())
                max_val = max(data_column.values())
        self.model.map_model.min_val = min_val
        self.model.map_model.max_val = max_val
        # set selected column
        self.model.selected_column = value

    @Slot()
    def get_column_names(self):
        if self.model.dataset_dir is None:
            return []
        columns_names = set()
        for feature in self.model.data["features"]:
            columns_names = columns_names | set(feature["properties"].keys())
        columns_names -= set(["track_id"])
        return sorted(list(columns_names))

    @Slot()
    def get_selected_column(self):
        if self.model.selected_column is None:
            return {}
        columns_names = self.get_column_names()
        try:
            column = columns_names[self.model.selected_column]
        except IndexError:
            return {}
        else:
            return self.get_column(column)

    @Slot()
    def get_column(self, column):
        if self.model.dataset_dir is None:
            return {}
        column_values = {}
        for feature in self.model.data["features"]:
            track_id = feature["properties"]["track_id"]
            try:
                val = feature["properties"][column]
                if val is None:
                    val = np.nan
                column_values[track_id] = val
            except KeyError:
                continue
        return column_values

    def update_dataset_stats(self):
        if self.model.dataset_dir is None:
            return

        self.thread_dataset_stats = QThread()
        self.worker_dataset_stats = ComputeDatasetStatsWorker(
            self.model.dataset_dir,
            self.model.dataset_version,
            self.model.data
        )
        self.worker_dataset_stats.moveToThread(self.thread_dataset_stats)

        # connect signals and slots
        self.thread_dataset_stats.started.connect(self.worker_dataset_stats.run)
        self.worker_dataset_stats.finished.connect(self.thread_dataset_stats.quit)

        def worker_finished():
            if self.worker_dataset_stats is not None:
                self.worker_dataset_stats.deleteLater()
                self.worker_dataset_stats = None

        def thread_finished():
            if self.thread_dataset_stats is not None:
                self.thread_dataset_stats.deleteLater()
                self.thread_dataset_stats = None

        self.worker_dataset_stats.finished.connect(worker_finished)
        self.thread_dataset_stats.finished.connect(thread_finished)

        self.thread_dataset_stats.start()

        # update dataset stats when thread is finished
        self.worker_dataset_stats.finished.connect(lambda stats: setattr(self.model, "dataset_stats", stats))

    @Slot()
    def stop_background_threads(self):
        if self.thread_dataset_stats is not None and self.worker_dataset_stats is not None:
            self.worker_dataset_stats.is_cancelled = True
            self.thread_dataset_stats.quit()
            self.thread_dataset_stats.wait()         
            self.thread_dataset_stats.deleteLater()
            self.thread_dataset_stats = None
            self.worker_dataset_stats = None



class ComputeDatasetStatsWorker(QObject):
    finished = Signal(object)

    def __init__(self, dataset_dir, dataset_version, data):
        super().__init__()
        self.is_cancelled = False
        self.dataset_dir = dataset_dir
        self.dataset_version = dataset_version
        self.data = data
    
    def run(self):
        print("Started background thread")
        # num modules
        num_modules = 0
        for feature in self.data["features"]:
            if self.is_cancelled:
                #self.finished.emit(None)
                print("cancelled thread")
                return
            geometry = feature["geometry"]["type"]
            if geometry == "Polygon":
                num_modules += 1

        print("Got num modules")

        # num patches
        num_patches = 0
        if self.dataset_version == "v1":
            patches_dir = os.path.join(self.dataset_dir, "patches_final", "radiometric")
        elif self.dataset_version == "v2":
            patches_dir = os.path.join(self.dataset_dir, "patches", "radiometric")
        for _, _, files in os.walk(patches_dir):
            if self.is_cancelled:
                #self.finished.emit(None)
                print("cancelled thread")
                return
            num_patches += len(files)

        print("Got num patches")

        # flight duration
        timestamps = []
        with open(os.path.join(self.dataset_dir, "splitted", "timestamps.csv"), newline='') as csvfile:
            csvreader = csv.reader(csvfile, delimiter=',', quotechar='|')
            for row in csvreader:
                if self.is_cancelled:
                    #self.finished.emit(None)
                    print("cancelled thread")
                    return
                timestamps.append(datetime.datetime.fromisoformat(*row))                
        dt = (timestamps[-1] - timestamps[0]).total_seconds()
        hours, remainder = divmod(dt, 3600)
        minutes, seconds = divmod(remainder, 60)
        flight_duration = "{:02d}:{:02d}:{:0.3f}".format(int(hours), int(minutes), seconds)

        print("Got flight duration")

        # trajectory length
        pose_graph = pickle.load(open(os.path.join(self.dataset_dir, "mapping", "pose_graph.pkl"), "rb"))
        positions = []
        for node_id, data in pose_graph.nodes(data=True):
            if self.is_cancelled:
                #self.finished.emit(None)
                print("cancelled thread")
                return
            pose = data["pose"][3:].reshape(3, 1)
            positions.append(pose)
        trajectory_length = 0
        for i in range(1, len(positions)):
            if self.is_cancelled:
                #self.finished.emit(None)
                print("cancelled thread")
                return
            trajectory_length += np.linalg.norm(positions[i] - positions[i-1])

        print("Got trajectory length")

        print("Finished computing dataset stats")    
        stats = {
            "num_modules": num_modules,
            "num_patches": num_patches,
            "flight_duration": flight_duration,
            "trajectory_length": trajectory_length,
        }
        self.finished.emit(stats)



class MainModel(QObject):
    dataset_opened = Signal()
    dataset_closed = Signal()
    source_names_changed = Signal(object)
    selected_source_changed = Signal(str)
    selected_column_changed = Signal(int)
    track_id_changed = Signal(str, str)
    meta_changed = Signal()
    sun_reflections_changed = Signal(object)
    dataset_stats_changed = Signal()
    app_mode_changed = Signal(str)

    def __init__(self):
        super().__init__()
        self.dataset_dir = None
        self.dataset_version = None
        self.data = None
        self._meta = None
        self._sun_reflections = None
        self.patch_meta = None
        self.track_ids = None
        self._app_mode = None # "None", "data_visualization", "defect_annotation", "string_annotation"
        self._source_names = None
        self._dataset_is_open = False
        self._selected_source = None
        self._selected_column = None
        self._track_id = None
        self._dataset_stats = None

    @property
    def meta(self):
        return self._meta

    @meta.setter
    def meta(self, value):
        self._meta = value
        self.meta_changed.emit()

    @property
    def app_mode(self):
        return self._app_mode

    @app_mode.setter
    def app_mode(self, value):
        self._app_mode = value
        self.app_mode_changed.emit(value)

    @property
    def selected_column(self):
        return self._selected_column

    @selected_column.setter
    def selected_column(self, value):
        self._selected_column = value
        self.selected_column_changed.emit(value)

    @property
    def track_id(self):
        return self._track_id

    @track_id.setter
    def track_id(self, value):
        track_id_prev = self._track_id
        self._track_id = value
        self.track_id_changed.emit(track_id_prev, value)

    @property
    def dataset_is_open(self):
        return self._dataset_is_open

    @dataset_is_open.setter
    def dataset_is_open(self, value):
        self._dataset_is_open = value
        if self._dataset_is_open:
            self.dataset_opened.emit()
        else:
            self.dataset_closed.emit()

    @property
    def source_names(self):
        return self._source_names

    @source_names.setter
    def source_names(self, value):
        self._source_names = value
        self.source_names_changed.emit(value)

    @property
    def selected_source(self):
        return self._selected_source

    @selected_source.setter
    def selected_source(self, value):
        self._selected_source = value
        self.selected_source_changed.emit(value)

    @property
    def dataset_stats(self):
        return self._dataset_stats

    @dataset_stats.setter
    def dataset_stats(self, value):
        self._dataset_stats = value
        self.dataset_stats_changed.emit()

    @property
    def sun_reflections(self):
        return self._sun_reflections

    @sun_reflections.setter
    def sun_reflections(self, value):
        self._sun_reflections = value
        self.sun_reflections_changed.emit(value)


import os
import json
import copy

from PySide6.QtWidgets import QWidget, QMessageBox, QFileDialog
from PySide6.QtCore import Slot, Signal, QObject
from PySide6.QtGui import QIcon

from ..ui.ui_string_editor import Ui_StringEditor


class StringEditorView(QWidget):
    def __init__(self, model, controller, parent=None):
        super().__init__(parent)
        self.model = model
        self.controller = controller
        self.ui = Ui_StringEditor()
        self.ui.setupUi(self)
        self.parent = parent
        # setup icons
        self.ui.pushButtonNewString.setIcon(QIcon.fromTheme("list-add"))
        self.ui.pushButtonDeleteString.setIcon(QIcon.fromTheme("list-remove"))
        self.disable()

        # connect signals and slots
        self.ui.pushButtonNewString.clicked.connect(self.new_string)
        self.ui.pushButtonCancelString.clicked.connect(self.cancel_string)
        self.ui.pushButtonConfirmString.clicked.connect(self.confirm_string)
        self.ui.pushButtonStartDrawing.clicked.connect(self.start_drawing)
        self.ui.pushButtonPauseDrawing.clicked.connect(self.pause_drawing)
        self.ui.pushButtonEndDrawing.clicked.connect(self.end_drawing)
        self.ui.pushButtonDeleteString.clicked.connect(self.controller.string_editor_controller.delete_string)
        self.ui.lineEditTrackerID.textChanged.connect(lambda value: setattr(self.model.string_editor_model, 'tracker_id', value))
        self.model.string_editor_model.tracker_id_changed.connect(self.ui.lineEditTrackerID.setText)
        self.ui.lineEditArrayID.textChanged.connect(lambda value: setattr(self.model.string_editor_model, 'array_id', value))
        self.model.string_editor_model.array_id_changed.connect(self.ui.lineEditArrayID.setText)
        self.ui.lineEditInverterID.textChanged.connect(lambda value: setattr(self.model.string_editor_model, 'inverter_id', value))
        self.model.string_editor_model.inverter_id_changed.connect(self.ui.lineEditInverterID.setText)
        self.ui.lineEditStringID.textChanged.connect(lambda value: setattr(self.model.string_editor_model, 'string_id', value))
        self.model.string_editor_model.string_id_changed.connect(self.ui.lineEditStringID.setText)
        self.model.string_editor_model.drawing_paused_changed.connect(self.drawing_paused_changed)
        self.controller.string_editor_controller.show_validation_error.connect(self.show_validation_error)
        self.controller.string_editor_controller.confirm_string.connect(self.disable)
        self.controller.export_string_annotation.connect(self.controller.string_editor_controller.export_string_annotation)
        self.model.string_editor_model.selected_string_id_changed.connect(self.selected_string_id_changed)
        self.model.string_editor_model.temporary_string_data_changed.connect(self.temporary_string_data_changed)
        self.model.dataset_opened.connect(self.controller.string_editor_controller.set_default_values)
        self.model.dataset_opened.connect(lambda: self.ui.pushButtonNewString.setEnabled(True))
        self.model.dataset_closed.connect(self.close_string_annotation)
        self.controller.close_string_annotation.connect(self.close_string_annotation)

    def disable(self):
        self.ui.pushButtonStartDrawing.setEnabled(False)
        self.ui.pushButtonPauseDrawing.setEnabled(False)
        self.ui.pushButtonEndDrawing.setEnabled(False)
        self.ui.pushButtonCancelString.setEnabled(False)
        self.ui.pushButtonConfirmString.setEnabled(False)
        self.ui.pushButtonDeleteString.setEnabled(False)
        self.ui.lineEditTrackerID.setEnabled(False)
        self.ui.lineEditArrayID.setEnabled(False)
        self.ui.lineEditInverterID.setEnabled(False)
        self.ui.lineEditStringID.setEnabled(False)

    def start_drawing(self):
        self.controller.string_editor_controller.drawing_started.emit()
        self.ui.pushButtonStartDrawing.setEnabled(False)
        self.ui.pushButtonPauseDrawing.setEnabled(True)
        self.ui.pushButtonEndDrawing.setEnabled(True)

    def pause_drawing(self):
        self.model.string_editor_model.drawing_paused = not self.model.string_editor_model.drawing_paused

    def end_drawing(self):
        self.controller.string_editor_controller.drawing_ended.emit()

    def new_string(self):
        self.ui.pushButtonNewString.setEnabled(False)
        self.controller.string_editor_controller.reset_temp_string_data()
        self.controller.string_editor_controller.new_string.emit()
        self.ui.pushButtonStartDrawing.setEnabled(True)
        self.ui.pushButtonPauseDrawing.setEnabled(False)
        self.ui.pushButtonEndDrawing.setEnabled(False)
        self.ui.pushButtonCancelString.setEnabled(True)
    
    def cancel_string(self):
        self.controller.string_editor_controller.reset_temp_string_data()
        self.controller.string_editor_controller.cancel_string.emit()
        self.ui.pushButtonNewString.setEnabled(True)
        self.disable()

    def confirm_string(self):
        self.controller.string_editor_controller.update_string_annotation_data()

    @Slot()
    def temporary_string_data_changed(self):
        temporary_string_data = self.model.string_editor_model.temporary_string_data
        if temporary_string_data is None:
            self.disable()
            self.ui.pushButtonNewString.setEnabled(True)
            return
        if len(temporary_string_data["modules"]) == 0:
            self.disable()
            self.ui.pushButtonNewString.setEnabled(True)
            msg = QMessageBox()
            msg.setWindowTitle("Error")
            msg.setText("No modules selected. Please click on the map to select modules after clicking \"Start Drawing\".")
            msg.setIcon(QMessageBox.Critical)
            msg.exec()
            return
        self.ui.pushButtonStartDrawing.setEnabled(False)
        self.ui.pushButtonPauseDrawing.setEnabled(False)
        self.ui.pushButtonEndDrawing.setEnabled(False)
        self.ui.pushButtonConfirmString.setEnabled(True)
        self.ui.lineEditTrackerID.setEnabled(True)
        self.ui.lineEditArrayID.setEnabled(True)
        self.ui.lineEditInverterID.setEnabled(True)
        self.ui.lineEditStringID.setEnabled(True)

    @Slot()
    def selected_string_id_changed(self):
        if self.model.string_editor_model.selected_string_id is not None:
            self.ui.pushButtonDeleteString.setEnabled(True)
        else:
            self.ui.pushButtonDeleteString.setEnabled(False)

    @Slot()
    def drawing_paused_changed(self):
        if self.model.string_editor_model.drawing_paused:
            self.ui.pushButtonPauseDrawing.setText("Continue")
        else:
            self.ui.pushButtonPauseDrawing.setText("Pause")

    @Slot(str)
    def show_validation_error(self, text):
        print("validation error")
        msg = QMessageBox()
        msg.setWindowTitle("Error")
        msg.setText(text)
        msg.setIcon(QMessageBox.Critical)
        msg.exec()

    @Slot()
    def close_string_annotation(self):
        self.controller.string_editor_controller.reset_temp_string_data()
        self.controller.string_editor_controller.reset_string_annotation_data
        self.disable()



class StringEditorController(QObject):
    new_string = Signal()
    confirm_string = Signal()
    cancel_string = Signal()
    show_validation_error = Signal(str)
    selected_string_id_changed = Signal(str, str)
    drawing_started = Signal()
    drawing_paused_changed = Signal(bool)
    drawing_ended = Signal()
    string_annotation_data_changed = Signal()

    def __init__(self, model):
        super().__init__()
        self.model = model
        # connect signals and slots
        self.model.dataset_opened.connect(self.load_annotation_file)
        self.model.string_editor_model.selected_string_id_changed.connect(self.selected_string_id_changed)
        self.model.string_editor_model.string_annotation_data_changed.connect(self.string_annotation_data_changed)
        self.model.string_editor_model.drawing_paused_changed.connect(self.drawing_paused_changed)

    def set_default_values(self):
        self.model.string_editor_model.tracker_id = "00"
        self.model.string_editor_model.array_id = "00"
        self.model.string_editor_model.inverter_id = "00"
        self.model.string_editor_model.string_id = "00"

    @Slot()
    def reset_temp_string_data(self):
        self.cancel_string.emit()
        self.model.string_editor_model.selected_string_id = None
        self.model.string_editor_model.temporary_string_data = None
        self.model.string_editor_model.drawing_paused = False

    @Slot()
    def reset_string_annotation_data(self):
        self.model.string_editor_model.string_annotation_data = None

    def is_valid(self, value):
        return isinstance(value, str) and value.isalnum() and (len(value) == 2 or len(value) == 3)

    def validate_string_id(self):
        # check if individual IDs are valid
        def error_msg(text):
            return "{} is invalid. Provide a 2-digit or 3-digit alphanumeric value.".format(text)

        if not self.is_valid(self.model.string_editor_model.tracker_id):
            self.show_validation_error.emit(error_msg("Tracker ID"))
            return False
        if not self.is_valid(self.model.string_editor_model.array_id):
            self.show_validation_error.emit(error_msg("Array ID"))
            return False
        if not self.is_valid(self.model.string_editor_model.inverter_id):
            self.show_validation_error.emit(error_msg("Inverter ID"))
            return False
        if not self.is_valid(self.model.string_editor_model.string_id):
            self.show_validation_error.emit(error_msg("String ID"))
            return False

        # check if a string with this id exists already
        string_id = "{}_{}_{}_{}".format(
            self.model.string_editor_model.tracker_id,
            self.model.string_editor_model.array_id,
            self.model.string_editor_model.inverter_id,
            self.model.string_editor_model.string_id,
        )
        if ((self.model.string_editor_model.string_annotation_data is not None) and
             string_id in self.model.string_editor_model.string_annotation_data["string_data"]):
            self.show_validation_error.emit("A string with this ID exists already.")
            return False

        return True

    def delete_string(self):
        selected_string_id = self.model.string_editor_model.selected_string_id
        if selected_string_id is None:
            return
        data = copy.deepcopy(self.model.string_editor_model.string_annotation_data)
        print(data)
        try:
            del data["string_data"][selected_string_id]
            
            for i in range(len(data["plant_id_track_id_mapping"])-1, -1, -1):  # iterate backwards
                plant_id, track_id = data["plant_id_track_id_mapping"][i]
                string_id = "_".join(str.split(plant_id, "_")[:-1])
                print(string_id, selected_string_id)
                if string_id == selected_string_id:
                    del data["plant_id_track_id_mapping"][i]

            print("Deleted string annotation data for string {}".format(selected_string_id))
        except KeyError:
            print("Failed to delete string annotation data for string {}".format(selected_string_id))
        else:
            self.model.string_editor_model.string_annotation_data = data
            self.save_annotation_file()
        self.model.string_editor_model.selected_string_id = None

    def save_annotation_file(self):
        print("Saving string annotation to file")
        if not self.model.dataset_is_open:
            return
        data = self.model.string_editor_model.string_annotation_data
        if data is None:
            return
        save_dir = os.path.join(self.model.dataset_dir, "annotations")
        os.makedirs(save_dir, exist_ok=True)
        json.dump(data, open(os.path.join(save_dir, "string_anotation.json"), "w"))

    @Slot()
    def load_annotation_file(self):
        print("Loading string annotation from file")
        if not self.model.dataset_is_open:
            return
        save_dir = os.path.join(self.model.dataset_dir, "annotations")
        try:
            data = json.load(open(os.path.join(save_dir, "string_anotation.json"), "r"))
        except FileNotFoundError:
            pass
        else:
            self.model.string_editor_model.string_annotation_data = data

    @Slot()
    def export_string_annotation(self):
        """Export current string annotation data to a user selected file."""
        file_name, _ = QFileDialog.getSaveFileName(
            None, "Export String Annotation", "", "", "JSON Files (*.json)")
        if file_name == "":
            return
        
        # make sure filename has JSON extension
        file_name = ".".join([os.path.splitext(file_name)[0], "json"])

        data = self.model.string_editor_model.string_annotation_data
        if data is None:
            return
        
        print("Exporting string annotation to ", file_name)
        json.dump(data, open(file_name, "w"))

    @Slot(str)
    def set_selected_string_id(self, selected_string_id):
        self.model.string_editor_model.selected_string_id = json.loads(selected_string_id)

    @Slot(result=str)
    def get_string_annotation_data(self):
        """Retrieve string annotation data from model and send to JS backend."""
        if not self.model.dataset_is_open:
            return json.dumps(None)

        string_annotation_data = self.model.string_editor_model.string_annotation_data
        if string_annotation_data is None:
            return json.dumps(None)

        return json.dumps(string_annotation_data)

    @Slot(str)
    def set_temporary_string_annotation_data(self, temporary_string_data):
        """Insert current string annotation into string annotation data."""
        temporary_string_data = json.loads(temporary_string_data)
        self.model.string_editor_model.temporary_string_data = temporary_string_data

    def update_string_annotation_data(self):
        ret = self.validate_string_id()
        if ret:
            temporary_string_data = self.model.string_editor_model.temporary_string_data
            modules = temporary_string_data["modules"]
            points = temporary_string_data["points"]
            paused = temporary_string_data["paused"]
            string_id = "{}_{}_{}_{}".format(
                self.model.string_editor_model.tracker_id,
                self.model.string_editor_model.array_id,
                self.model.string_editor_model.inverter_id,
                self.model.string_editor_model.string_id,
            )

            data = copy.deepcopy(self.model.string_editor_model.string_annotation_data)
            if data is None:
                data = {
                    "string_data": {},
                    "plant_id_track_id_mapping": []
                }

            # insert data of current string (e.g. points, contained modules)
            data["string_data"][string_id] = {
                "track_ids": [module["track_id"] for module in modules],
                "points": points,
                "paused": paused
            }

            # update plant_id / track_id mapping
            for module_id, module in enumerate(modules):
                plant_id = "{}_{:02d}".format(string_id, module_id)
                data["plant_id_track_id_mapping"].append((plant_id, module["track_id"]))

            self.model.string_editor_model.string_annotation_data = data
            self.model.string_editor_model.temporary_string_data = None
            self.reset_temp_string_data()
            self.save_annotation_file()
            self.confirm_string.emit()



class StringEditorModel(QObject):
    tracker_id_changed = Signal(str)
    array_id_changed = Signal(str)
    inverter_id_changed = Signal(str)
    string_id_changed = Signal(str)
    selected_string_id_changed = Signal(str, str)
    drawing_paused_changed = Signal(bool)
    temporary_string_data_changed = Signal()
    string_annotation_data_changed = Signal()

    def __init__(self):
        super().__init__()
        self._tracker_id = ""
        self._array_id = ""
        self._inverter_id = ""
        self._string_id = ""
        self._selected_string_id = None
        self._drawing_paused = False
        self._temporary_string_data = None
        self._string_annotation_data = None  # string data of entire plant

    @property
    def tracker_id(self):
        return self._tracker_id

    @tracker_id.setter
    def tracker_id(self, value):
        self._tracker_id = value
        self.tracker_id_changed.emit(value)

    @property
    def array_id(self):
        return self._array_id

    @array_id.setter
    def array_id(self, value):
        self._array_id = value
        self.array_id_changed.emit(value)

    @property
    def inverter_id(self):
        return self._inverter_id

    @inverter_id.setter
    def inverter_id(self, value):
        self._inverter_id = value
        self.inverter_id_changed.emit(value)

    @property
    def string_id(self):
        return self._string_id

    @string_id.setter
    def string_id(self, value):
        self._string_id = value
        self.string_id_changed.emit(value)

    @property
    def selected_string_id(self):
        return self._selected_string_id

    @selected_string_id.setter
    def selected_string_id(self, value):
        selected_string_id_prev = self._selected_string_id
        self._selected_string_id = value
        self.selected_string_id_changed.emit(selected_string_id_prev, value)

    @property
    def drawing_paused(self):
        return self._drawing_paused

    @drawing_paused.setter
    def drawing_paused(self, value):
        self._drawing_paused = value
        self.drawing_paused_changed.emit(value)

    @property
    def string_annotation_data(self):
        return self._string_annotation_data

    @string_annotation_data.setter
    def string_annotation_data(self, value):
        self._string_annotation_data = value
        self.string_annotation_data_changed.emit()

    @property
    def temporary_string_data(self):
        return self._temporary_string_data

    @temporary_string_data.setter
    def temporary_string_data(self, value):
        self._temporary_string_data = value
        self.temporary_string_data_changed.emit()
        
       
        


import os
import json
import pkg_resources

from PySide6.QtWidgets import QWidget, QCheckBox, QSpacerItem, QSizePolicy, \
    QMessageBox, QFileDialog
from PySide6.QtCore import Slot, Signal, QObject

from ..ui.ui_annotation_editor import Ui_AnnotationEditor


class AnnotationEditorView(QWidget):
    def __init__(self, model, controller, parent=None):
        super().__init__(parent)
        self.model = model
        self.controller = controller
        self.ui = Ui_AnnotationEditor()
        self.ui.setupUi(self)
        self.loadDefectsScheme()
        # connect signals and slots
        self.model.track_id_changed.connect(self.enable_disable)
        self.model.app_mode_changed.connect(self.enable_disable)
        self.model.track_id_changed.connect(self.update_checkbox_states)
        self.model.annotation_editor_model.annotation_data_changed.connect(self.update_checkbox_states)

        self.controller.new_defect_annotation.connect(self.controller.annotation_editor_controller.set_annotation_data)
        self.controller.save_defect_annotation.connect(self.controller.annotation_editor_controller.save_annotation_file)
        self.controller.load_defect_annotation.connect(self.controller.annotation_editor_controller.load_annotation_file)
        self.controller.close_defect_annotation.connect(self.controller.annotation_editor_controller.close_annotation)
        self.controller.mainwindow_close_requested.connect(self.controller.annotation_editor_controller.mainwindow_close_requested)
        self.controller.dataset_close_requested.connect(self.controller.annotation_editor_controller.dataset_close_requested)

    def loadDefectsScheme(self):
        try:
            defects_scheme = json.load(open(pkg_resources.resource_filename("src.resources", "defect_schema.json"), "r"))
        except FileNotFoundError:
            pass
        else:
            self.ui.checkboxes = []
            for defect in defects_scheme:
                checkbox = QCheckBox("{} - {}".format(defect["name"], defect["description"]))
                checkbox.setObjectName(u"defect_checkbox_{}".format(defect["name"]))
                checkbox.setToolTip(", ".join(defect["examples"]))
                checkbox.stateChanged.connect(self.controller.annotation_editor_controller.update_annotation_data)
                checkbox.stateChanged.connect(lambda: setattr(self.model.annotation_editor_model, 'has_changes', True))
                checkbox.setEnabled(False)
                self.ui.scrollAreaWidgetContents.layout().addWidget(checkbox)
                self.ui.checkboxes.append(checkbox)
            self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)
            self.ui.scrollAreaWidgetContents.layout().addItem(self.verticalSpacer)

    @Slot()
    def enable_disable(self):
        if self.model.app_mode != "defect_annotation":
            for checkbox in self.ui.checkboxes:
                checkbox.setEnabled(False)
            return
        if self.model.track_id is None:
            for checkbox in self.ui.checkboxes:
                checkbox.setEnabled(False)
            return
        for checkbox in self.ui.checkboxes:
                checkbox.setEnabled(True)

    @Slot()
    def update_checkbox_states(self):
        """Update checkboxes based on annotation state"""
        # clear checkboxes if dataset is closed
        if self.model.annotation_editor_model.annotation_data is None:
            for checkbox in self.ui.checkboxes:
                if checkbox.isChecked():
                    checkbox.setChecked(False)
            return

        track_id = self.model.track_id
        if track_id is None:
            for checkbox in self.ui.checkboxes:
                if checkbox.isChecked():
                    checkbox.setChecked(False)
            return

        defects = self.model.annotation_editor_model.annotation_data[track_id]
        for checkbox in self.ui.checkboxes:
            checkbox_name = checkbox.objectName()
            defect = checkbox_name[16:]
            if defect in defects:
                if not checkbox.isChecked():
                    checkbox.setChecked(True)
            else:
                if checkbox.isChecked():
                    checkbox.setChecked(False)



class AnnotationEditorController(QObject):
    close_dataset = Signal()

    def __init__(self, model):
        super().__init__()
        self.model = model

    @Slot()
    def set_annotation_data(self):
        self.model.annotation_editor_model.annotation_data = {track_id: [] for track_id in self.model.track_ids}
        self.model.annotation_editor_model.current_file_name = "new annotation.csv"
        self.model.annotation_editor_model.has_changes = False

    @Slot()
    def reset_annotation_data(self):
        self.model.annotation_editor_model.annotation_data = None
        self.model.annotation_editor_model.current_file_name = "new annotation.csv"
        self.model.annotation_editor_model.has_changes = False

    @Slot()
    def update_annotation_data(self, value):
        track_id = self.model.track_id
        print("Updating annotation data ", self.model.dataset_is_open, self.model.track_id, (self.model.annotation_editor_model.annotation_data is None))
        if not self.model.dataset_is_open:
            return
        if self.model.annotation_editor_model.annotation_data is None:
            return
        if track_id is None:
            return

        checkbox = self.sender()
        checkbox_name = checkbox.objectName()
        checked = value != 0
        defect = checkbox_name[16:]

        # update model
        if checked:
            print("checked: ", checkbox_name)
            if defect not in self.model.annotation_editor_model.annotation_data[track_id]:
                self.model.annotation_editor_model.annotation_data[track_id].append(defect)
                self.model.annotation_editor_model.annotation_data_changed.emit()  # won't be fired automatically inside model because we only change an internal value

        elif not checked:
            print("unchecked: ", checkbox_name)
            if defect in self.model.annotation_editor_model.annotation_data[track_id]:
                self.model.annotation_editor_model.annotation_data[track_id].remove(defect)
                self.model.annotation_editor_model.annotation_data_changed.emit()

        self.print_annotation_data()

    @Slot(object)
    def mainwindow_close_requested(self, event):
        if self.model.app_mode == "defect_annotation":
            status = self.save_changes_dialog()
            if status == "no_changes" or status == "saved" or status == "discarded":
                event.accept()
            elif status == "cancelled":
                event.ignore()

    @Slot()
    def dataset_close_requested(self):
        status = self.save_changes_dialog()
        if status == "no_changes" or status == "saved" or status == "discarded":
            self.reset_annotation_data()
            self.close_dataset.emit()
            self.model.app_mode = "data_visualization"       
        elif status == "cancelled":
            return

    @Slot()
    def close_annotation(self):
        status = self.save_changes_dialog()
        if status == "no_changes" or status == "saved" or status == "discarded":
            self.reset_annotation_data()
            self.model.app_mode = "data_visualization"
        elif status == "cancelled":
            return

    @Slot()
    def save_changes_dialog(self):
        if not self.model.annotation_editor_model.has_changes:
            return "no_changes"
        msg = QMessageBox()
        msg.setWindowTitle("Save changes?")
        msg.setText("Do you want to save unchanged changes to the defect annotation?")
        msg.setIcon(QMessageBox.Question)
        msg.setStandardButtons(QMessageBox.Yes|QMessageBox.No|QMessageBox.Cancel)
        answer = msg.exec() 
        if answer == QMessageBox.Yes:
            ret = self.save_annotation_file()
            return ret

        elif answer == QMessageBox.No:
            print("Discarding defect annotation")
            self.reset_annotation_data()
            return "discarded"

        elif answer == QMessageBox.Cancel:
            print("Cancelled saving defect annotation")
            return "cancelled"

    @Slot()
    def save_annotation_file(self):
        default_file_name = self.model.annotation_editor_model.current_file_name
        file_name, _ = QFileDialog.getSaveFileName(
            None, "Save Defect Annotation", default_file_name, "", "JSON Files (*.json)")
        if file_name == "":
            return "cancelled"
        
        # make sure filename has JSON extension
        file_name = ".".join([os.path.splitext(file_name)[0], "json"])

        if self.model.annotation_editor_model.annotation_data is None:
            return "cancelled"
        
        print("Saving to ", file_name)
        annotation_data_json = self.annotation_data_to_json(self.model.annotation_editor_model.annotation_data)
        json.dump(annotation_data_json, open(file_name, "w"))

        self.model.annotation_editor_model.current_file_name = file_name
        self.model.annotation_editor_model.has_changes = False
        return "saved"

    @Slot()
    def load_annotation_file(self):
        file_name, _ = QFileDialog.getOpenFileName(
            None, "Open Defect Annotation", "", "JSON Files (*.json)", "")
        if file_name == "":
            return
        if self.model.track_ids is None:
            return
        try:
            annotation_data_json = json.load(open(file_name, "r"))
        except json.decoder.JSONDecodeError:
            msg = QMessageBox()
            msg.setWindowTitle("Error")
            msg.setText("Not a valid defect annotation file.")
            msg.setIcon(QMessageBox.Critical)
            msg.exec()
            return

        # convert to internal format of annotation data
        try:
            data = self.annotation_data_from_json(annotation_data_json)
        except (KeyError, IndexError):
            msg = QMessageBox()
            msg.setWindowTitle("Error")
            msg.setText("Not a valid defect annotation file.")
            msg.setIcon(QMessageBox.Critical)
            msg.exec()
            return

        # validate that this defect annotation belongs to the opened dataset
        if not set(data.keys()) == set(self.model.track_ids):
            msg = QMessageBox()
            msg.setWindowTitle("Error")
            msg.setText("This defect annotation file does not belong to the currently opened dataset. Please select another file.")
            msg.setIcon(QMessageBox.Critical)
            msg.exec()
            return

        self.model.annotation_editor_model.annotation_data = data
        self.model.annotation_editor_model.current_file_name = file_name
        self.model.app_mode = "defect_annotation"
        print("Successfully loaded defect annotations")

    def annotation_data_to_json(self, data):
        data_json = [{"plant_id": track_id, "faults": defects} for track_id, defects in data.items()]
        return data_json

    def annotation_data_from_json(self, data_json):
        data = {}
        for row in data_json:
            track_id = row["plant_id"]
            defects = row["faults"]
            data[track_id] = defects
        return data

    # for debugging
    def print_annotation_data(self):
        for track_id, defects in self.model.annotation_editor_model.annotation_data.items():
            if len(defects) > 0:
                print(track_id, defects)



class AnnotationEditorModel(QObject):
    annotation_data_changed = Signal()
    has_changes_changed = Signal(bool)
    current_file_name_changed = Signal(str)

    def __init__(self):
        super().__init__()
        self._annotation_data = None
        self._has_changes = False  # whether there are unsaved changes
        self._current_file_name = "new annotation.csv"
    
    @property
    def annotation_data(self):
        return self._annotation_data

    @annotation_data.setter
    def annotation_data(self, value):
        self._annotation_data = value
        self.annotation_data_changed.emit()

    @property
    def has_changes(self):
        return self._has_changes

    @has_changes.setter
    def has_changes(self, value):
        self._has_changes = value
        self.has_changes_changed.emit(value)

    @property
    def current_file_name(self):
        return self._current_file_name

    @current_file_name.setter
    def current_file_name(self, value):
        self._current_file_name = value
        self.current_file_name_changed.emit(value)


# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'analysis_details.ui'
##
## Created by: Qt User Interface Compiler version 6.2.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QFormLayout, QHBoxLayout, QLabel,
    QLineEdit, QListWidget, QListWidgetItem, QPushButton,
    QSizePolicy, QSpacerItem, QWidget)

class Ui_AnalysisDetails(object):
    def setupUi(self, AnalysisDetails):
        if not AnalysisDetails.objectName():
            AnalysisDetails.setObjectName(u"AnalysisDetails")
        AnalysisDetails.resize(625, 336)
        sizePolicy = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(AnalysisDetails.sizePolicy().hasHeightForWidth())
        AnalysisDetails.setSizePolicy(sizePolicy)
        self.formLayout = QFormLayout(AnalysisDetails)
        self.formLayout.setObjectName(u"formLayout")
        self.label = QLabel(AnalysisDetails)
        self.label.setObjectName(u"label")

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.label)

        self.lineEditAnalysisType = QLineEdit(AnalysisDetails)
        self.lineEditAnalysisType.setObjectName(u"lineEditAnalysisType")
        self.lineEditAnalysisType.setReadOnly(True)

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self.lineEditAnalysisType)

        self.label_2 = QLabel(AnalysisDetails)
        self.label_2.setObjectName(u"label_2")

        self.formLayout.setWidget(1, QFormLayout.LabelRole, self.label_2)

        self.lineEditTimestamp = QLineEdit(AnalysisDetails)
        self.lineEditTimestamp.setObjectName(u"lineEditTimestamp")
        self.lineEditTimestamp.setReadOnly(True)

        self.formLayout.setWidget(1, QFormLayout.FieldRole, self.lineEditTimestamp)

        self.label_3 = QLabel(AnalysisDetails)
        self.label_3.setObjectName(u"label_3")

        self.formLayout.setWidget(2, QFormLayout.LabelRole, self.label_3)

        self.lineEditDatasetPath = QLineEdit(AnalysisDetails)
        self.lineEditDatasetPath.setObjectName(u"lineEditDatasetPath")
        self.lineEditDatasetPath.setReadOnly(True)

        self.formLayout.setWidget(2, QFormLayout.FieldRole, self.lineEditDatasetPath)

        self.label_4 = QLabel(AnalysisDetails)
        self.label_4.setObjectName(u"label_4")

        self.formLayout.setWidget(3, QFormLayout.LabelRole, self.label_4)

        self.listWidgetHyperparameters = QListWidget(AnalysisDetails)
        self.listWidgetHyperparameters.setObjectName(u"listWidgetHyperparameters")

        self.formLayout.setWidget(3, QFormLayout.FieldRole, self.listWidgetHyperparameters)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.pushButtonClose = QPushButton(AnalysisDetails)
        self.pushButtonClose.setObjectName(u"pushButtonClose")

        self.horizontalLayout.addWidget(self.pushButtonClose)


        self.formLayout.setLayout(4, QFormLayout.FieldRole, self.horizontalLayout)


        self.retranslateUi(AnalysisDetails)

        QMetaObject.connectSlotsByName(AnalysisDetails)
    # setupUi

    def retranslateUi(self, AnalysisDetails):
        AnalysisDetails.setWindowTitle(QCoreApplication.translate("AnalysisDetails", u"Analysis Details", None))
        self.label.setText(QCoreApplication.translate("AnalysisDetails", u"Analysis Type:", None))
        self.label_2.setText(QCoreApplication.translate("AnalysisDetails", u"Timestamp:", None))
        self.label_3.setText(QCoreApplication.translate("AnalysisDetails", u"Dataset Path:", None))
        self.label_4.setText(QCoreApplication.translate("AnalysisDetails", u"Hyperparameters:", None))
        self.pushButtonClose.setText(QCoreApplication.translate("AnalysisDetails", u"Close", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'source_frame.ui'
##
## Created by: Qt User Interface Compiler version 6.2.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QFrame, QGridLayout,
    QHBoxLayout, QLabel, QSizePolicy, QSpacerItem,
    QSpinBox, QWidget)

class Ui_SourceFrame(object):
    def setupUi(self, SourceFrame):
        if not SourceFrame.objectName():
            SourceFrame.setObjectName(u"SourceFrame")
        SourceFrame.resize(675, 455)
        self.gridLayout = QGridLayout(SourceFrame)
        self.gridLayout.setObjectName(u"gridLayout")
        self.sourceFrameLabel = QLabel(SourceFrame)
        self.sourceFrameLabel.setObjectName(u"sourceFrameLabel")
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sourceFrameLabel.sizePolicy().hasHeightForWidth())
        self.sourceFrameLabel.setSizePolicy(sizePolicy)
        self.sourceFrameLabel.setMinimumSize(QSize(50, 50))
        self.sourceFrameLabel.setScaledContents(False)

        self.gridLayout.addWidget(self.sourceFrameLabel, 1, 0, 1, 1)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.label_2 = QLabel(SourceFrame)
        self.label_2.setObjectName(u"label_2")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.label_2.sizePolicy().hasHeightForWidth())
        self.label_2.setSizePolicy(sizePolicy1)

        self.horizontalLayout.addWidget(self.label_2)

        self.minTempSpinBox = QSpinBox(SourceFrame)
        self.minTempSpinBox.setObjectName(u"minTempSpinBox")
        self.minTempSpinBox.setMinimum(-100)
        self.minTempSpinBox.setMaximum(200)
        self.minTempSpinBox.setValue(30)

        self.horizontalLayout.addWidget(self.minTempSpinBox)

        self.label_3 = QLabel(SourceFrame)
        self.label_3.setObjectName(u"label_3")
        sizePolicy1.setHeightForWidth(self.label_3.sizePolicy().hasHeightForWidth())
        self.label_3.setSizePolicy(sizePolicy1)

        self.horizontalLayout.addWidget(self.label_3)

        self.maxTempSpinBox = QSpinBox(SourceFrame)
        self.maxTempSpinBox.setObjectName(u"maxTempSpinBox")
        self.maxTempSpinBox.setMinimum(-100)
        self.maxTempSpinBox.setMaximum(200)
        self.maxTempSpinBox.setValue(50)

        self.horizontalLayout.addWidget(self.maxTempSpinBox)

        self.label_4 = QLabel(SourceFrame)
        self.label_4.setObjectName(u"label_4")
        sizePolicy1.setHeightForWidth(self.label_4.sizePolicy().hasHeightForWidth())
        self.label_4.setSizePolicy(sizePolicy1)

        self.horizontalLayout.addWidget(self.label_4)

        self.line = QFrame(SourceFrame)
        self.line.setObjectName(u"line")
        sizePolicy2 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Preferred)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.line.sizePolicy().hasHeightForWidth())
        self.line.setSizePolicy(sizePolicy2)
        self.line.setFrameShape(QFrame.VLine)
        self.line.setFrameShadow(QFrame.Sunken)

        self.horizontalLayout.addWidget(self.line)

        self.label_5 = QLabel(SourceFrame)
        self.label_5.setObjectName(u"label_5")
        sizePolicy1.setHeightForWidth(self.label_5.sizePolicy().hasHeightForWidth())
        self.label_5.setSizePolicy(sizePolicy1)

        self.horizontalLayout.addWidget(self.label_5)

        self.colormapComboBox = QComboBox(SourceFrame)
        self.colormapComboBox.setObjectName(u"colormapComboBox")

        self.horizontalLayout.addWidget(self.colormapComboBox)

        self.horizontalSpacer = QSpacerItem(0, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)


        self.gridLayout.addLayout(self.horizontalLayout, 0, 0, 1, 1)


        self.retranslateUi(SourceFrame)

        QMetaObject.connectSlotsByName(SourceFrame)
    # setupUi

    def retranslateUi(self, SourceFrame):
        SourceFrame.setWindowTitle(QCoreApplication.translate("SourceFrame", u"Form", None))
        self.sourceFrameLabel.setText("")
        self.label_2.setText(QCoreApplication.translate("SourceFrame", u"Temp Range", None))
        self.label_3.setText(QCoreApplication.translate("SourceFrame", u"-", None))
        self.label_4.setText(QCoreApplication.translate("SourceFrame", u"\u00b0C", None))
        self.label_5.setText(QCoreApplication.translate("SourceFrame", u"Colormap", None))
    # retranslateUi





import datetime

from PySide6.QtWidgets import QWidget, QMessageBox
from PySide6.QtCore import Qt, Slot, QThread, Slot, Signal, QObject

from ..ui.ui_analysis import Ui_Analysis
from ..analysis.temperatures import AnalysisModuleTemperaturesWorker
from ..analysis.sun_filter import AnalysisSunFilterWorker


class AnalysisView(QWidget):
    def __init__(self, model, controller, parent=None):
        super().__init__(parent)
        self.setWindowFlags(Qt.Window)  # open in new window
        self.model = model
        self.controller = controller
        self.parent = parent
        self.ui = Ui_Analysis()
        self.ui.setupUi(self)
        self.reset()
        # connect signals and slots
        self.model.dataset_closed.connect(self.close)
        self.ui.pushButtonCompute.clicked.connect(self.controller.analysis_controller.compute)
        self.ui.pushButtonCancel.clicked.connect(self.controller.analysis_controller.cancel)
        self.ui.pushButtonOk.clicked.connect(self.close)
        self.model.analysis_model.name_changed.connect(self.ui.nameLineEdit.setText)
        self.ui.nameLineEdit.textChanged.connect(lambda value: setattr(self.model.analysis_model, 'name', value))
        self.model.analysis_model.progress_changed.connect(self.ui.progressBar.setValue)
        self.model.analysis_model.progress_text_changed.connect(self.ui.progressLabel.setText)
        self.controller.analysis_controller.name_exists.connect(self.show_name_exist_dialog)
        self.model.analysis_model.status_changed.connect(self.status_changed)
        self.ui.tabWidget.currentChanged.connect(self.tabChanged)
        self.model.analysis_model.active_tab_widget_changed.connect(self.ui.tabWidget.setCurrentWidget)
        self.model.analysis_model.active_tab_widget_changed.connect(self.active_tab_widget_changed)

        # module temperatures
        self.model.analysis_model.module_temperatures.border_margin_changed.connect(self.ui.spinBoxTruncateWidth.setValue)
        self.ui.spinBoxTruncateWidth.valueChanged.connect(lambda value: setattr(self.model.analysis_model.module_temperatures, 'border_margin', value))
        self.model.analysis_model.module_temperatures.neighbor_radius_changed.connect(self.ui.spinBoxNeighborRadius.setValue)
        self.ui.spinBoxNeighborRadius.valueChanged.connect(lambda value: setattr(self.model.analysis_model.module_temperatures, 'neighbor_radius', value))
        self.model.analysis_model.module_temperatures.ignore_sun_reflections_changed.connect(self.ui.checkBoxIgnoreSunReflections.setChecked)
        self.ui.checkBoxIgnoreSunReflections.stateChanged.connect(lambda value: setattr(self.model.analysis_model.module_temperatures, 'ignore_sun_reflections', bool(value)))
        
        # sun filter
        self.model.analysis_model.sun_filter.threshold_temp_changed.connect(self.ui.spinBoxThresholdTemp.setValue)
        self.ui.spinBoxThresholdTemp.valueChanged.connect(lambda value: setattr(self.model.analysis_model.sun_filter, 'threshold_temp', value))
        self.model.analysis_model.sun_filter.threshold_loc_changed.connect(self.ui.spinBoxThresholdLoc.setValue)
        self.ui.spinBoxThresholdLoc.valueChanged.connect(lambda value: setattr(self.model.analysis_model.sun_filter, 'threshold_loc', value))
        self.model.analysis_model.sun_filter.threshold_changepoint_changed.connect(self.ui.spinBoxThresholdChangepoint.setValue)
        self.ui.spinBoxThresholdChangepoint.valueChanged.connect(lambda value: setattr(self.model.analysis_model.sun_filter, 'threshold_changepoint', value))
        self.model.analysis_model.sun_filter.segment_length_threshold_changed.connect(self.ui.spinBoxSegmentLengthThreshold.setValue)
        self.ui.spinBoxSegmentLengthThreshold.valueChanged.connect(lambda value: setattr(self.model.analysis_model.sun_filter, 'segment_length_threshold', value))

        # set default values
        self.controller.analysis_controller.reset()
        self.model.analysis_model.active_tab_widget = self.ui.tabSunFilter

    def reset(self):
        self.ui.pushButtonOk.hide()
        self.ui.pushButtonCompute.show()
        self.ui.pushButtonCancel.setEnabled(False)
        self.ui.pushButtonCompute.setEnabled(True)
        self.ui.spinBoxTruncateWidth.setEnabled(True)
        self.ui.spinBoxNeighborRadius.setEnabled(True)
        self.ui.spinBoxThresholdTemp.setEnabled(True)
        self.ui.spinBoxThresholdLoc.setEnabled(True)
        self.ui.spinBoxThresholdChangepoint.setEnabled(True)
        self.ui.spinBoxSegmentLengthThreshold.setEnabled(True)
        self.enable_disable_sun_reflections()

    @Slot(int)
    def tabChanged(self, idx):
        self.model.analysis_model.active_tab_widget = self.ui.tabWidget.widget(idx) 

    @Slot(object)
    def active_tab_widget_changed(self, widget):
        if widget.objectName() == "tabSunFilter":
            self.model.analysis_model.name = "Sun Filter"
            self.ui.nameLineEdit.setEnabled(False)
        else:
            time = datetime.datetime.utcnow().strftime("%Y-%m-%dT%H-%M-%S")
            self.model.analysis_model.name = "Analysis {}".format(time)
            self.ui.nameLineEdit.setEnabled(True)

    def enable_disable_sun_reflections(self):
        if self.model.sun_reflections is None:
            self.ui.checkBoxIgnoreSunReflections.setEnabled(False)
        else:
            self.ui.checkBoxIgnoreSunReflections.setEnabled(True)

    @Slot(object)
    def status_changed(self, status):
        if status is None:
            self.reset()
        elif status == "started":
            self.ui.pushButtonCancel.setEnabled(True)
            self.ui.pushButtonCompute.setEnabled(False)
            self.ui.spinBoxTruncateWidth.setEnabled(False)
            self.ui.spinBoxNeighborRadius.setEnabled(False)
            self.ui.checkBoxIgnoreSunReflections.setEnabled(False)
            self.ui.spinBoxThresholdTemp.setEnabled(False)
            self.ui.spinBoxThresholdLoc.setEnabled(False)
            self.ui.spinBoxThresholdChangepoint.setEnabled(False)
            self.ui.spinBoxSegmentLengthThreshold.setEnabled(False)
        elif status == "cancelled":
            self.ui.pushButtonCompute.hide()
            self.ui.pushButtonOk.show()
            self.ui.pushButtonCancel.setEnabled(False)
        elif status == "finished":
            self.ui.pushButtonCompute.hide()
            self.ui.pushButtonOk.show()
            self.ui.pushButtonCancel.setEnabled(False)
            self.controller.update_source_names()
            self.controller.load_sun_reflections()

    @Slot()
    def show_name_exist_dialog(self):
        msg = QMessageBox()
        msg.setWindowTitle("Error")
        msg.setText("An analysis with this name already exists. Please specify a different name.")
        msg.setIcon(QMessageBox.Critical)
        msg.exec()

    def closeEvent(self, event):
        self.controller.analysis_controller.cancel()
        event.accept()



class AnalysisController(QObject):
    name_exists = Signal()

    def __init__(self, model):
        super().__init__()
        self.model = model
        self.thread = None
        self.worker = None

    @Slot()
    def reset(self):
        self.model.analysis_model.progress = 0
        self.model.analysis_model.progress_text = ""
        self.model.analysis_model.status = None
        # module temperatures
        self.model.analysis_model.module_temperatures.border_margin = 5
        self.model.analysis_model.module_temperatures.neighbor_radius = 7
        self.model.analysis_model.module_temperatures.ignore_sun_reflections = False
        # sun filter
        self.model.analysis_model.sun_filter.threshold_temp = 5.0
        self.model.analysis_model.sun_filter.threshold_loc = 10.0
        self.model.analysis_model.sun_filter.threshold_changepoint = 10.0
        self.model.analysis_model.sun_filter.segment_length_threshold = 0.3

    @Slot()
    def compute(self):
        if not self.model.dataset_is_open:
            return
        
        if self.model.analysis_model.name in self.model.source_names:
            self.name_exists.emit()
            return

        if not self.model.analysis_model.active_tab_widget or self.model.analysis_model.active_tab_widget.objectName() is None:
            return

        self.model.analysis_model.status = "started"
        self.thread = QThread()

        if self.model.analysis_model.active_tab_widget.objectName() == "tabSunFilter":
            self.worker = AnalysisSunFilterWorker(
                self.model.dataset_dir, 
                self.model.dataset_version,
                self.model.analysis_model.name,
                self.model.dataset_settings_model.gain,
                self.model.dataset_settings_model.offset,
                self.model.analysis_model.sun_filter.threshold_temp, 
                self.model.analysis_model.sun_filter.threshold_loc,
                self.model.analysis_model.sun_filter.threshold_changepoint,
                self.model.analysis_model.sun_filter.segment_length_threshold)
                
        elif self.model.analysis_model.active_tab_widget.objectName() == "tabModuleTemperatures":
            self.worker = AnalysisModuleTemperaturesWorker(
                self.model.dataset_dir,
                self.model.dataset_version,
                self.model.analysis_model.name,
                self.model.dataset_settings_model.gain,
                self.model.dataset_settings_model.offset,
                self.model.analysis_model.module_temperatures.border_margin, 
                self.model.analysis_model.module_temperatures.neighbor_radius,
                self.model.analysis_model.module_temperatures.ignore_sun_reflections,
                self.model.sun_reflections)
        self.worker.moveToThread(self.thread)
        self.thread.started.connect(self.worker.run)
        self.worker.finished.connect(self.thread.quit)
        self.worker.finished.connect(self.worker.deleteLater)
        self.thread.finished.connect(self.thread.deleteLater)
        self.worker.progress.connect(self.report_progress)
        self.worker.finished.connect(lambda: setattr(self.model.analysis_model, 'status', 'finished'))
        self.thread.start()

    @Slot()
    def report_progress(self, progress, cancelled, description=None):
        self.model.analysis_model.progress = round(progress*100)
        self.model.analysis_model.progress_text = description

        if cancelled:
            self.model.analysis_model.status = "cancelled"
            print("Cancelled module temperature analysis")
    
    @Slot()
    def cancel(self):
        if self.thread is not None and self.worker is not None:
            self.worker.is_cancelled = True



class AnalysisModel(QObject):
    name_changed = Signal(str)
    progress_changed = Signal(int)
    progress_text_changed = Signal(str)
    status_changed = Signal(object)
    active_tab_widget_changed = Signal(object)    

    def __init__(self):
        super().__init__()
        self._name = None
        self._progress = None
        self._progress_text = None
        self._status = None
        self._active_tab_widget = None        
        self.module_temperatures = AnalysisModuleTemperaturesModel()
        self.sun_filter = AnalysisSunFilterModel()
        
    @property
    def progress(self):
        return self._progress

    @progress.setter
    def progress(self, value):
        self._progress = value
        self.progress_changed.emit(value)

    @property
    def progress_text(self):
        return self._progress_text

    @progress_text.setter
    def progress_text(self, value):
        self._progress_text = value
        self.progress_text_changed.emit(value)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        self.name_changed.emit(value)

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        self._status = value
        self.status_changed.emit(value)

    @property
    def active_tab_widget(self):
        return self._active_tab_widget

    @active_tab_widget.setter
    def active_tab_widget(self, value):
        self._active_tab_widget = value
        self.active_tab_widget_changed.emit(value)



class AnalysisModuleTemperaturesModel(QObject):
    border_margin_changed = Signal(int)
    neighbor_radius_changed = Signal(int)
    ignore_sun_reflections_changed = Signal(bool)

    def __init__(self):
        super().__init__()
        self._border_margin = None
        self._neighbor_radius = None
        self._ignore_sun_reflections = None

    @property
    def border_margin(self):
        return self._border_margin

    @border_margin.setter
    def border_margin(self, value):
        self._border_margin = value
        self.border_margin_changed.emit(value)

    @property
    def neighbor_radius(self):
        return self._neighbor_radius

    @neighbor_radius.setter
    def neighbor_radius(self, value):
        self._neighbor_radius = value
        self.neighbor_radius_changed.emit(value)

    @property
    def ignore_sun_reflections(self):
        return self._ignore_sun_reflections

    @ignore_sun_reflections.setter
    def ignore_sun_reflections(self, value):
        self._ignore_sun_reflections = value
        self.ignore_sun_reflections_changed.emit(value)



class AnalysisSunFilterModel(QObject):
    threshold_temp_changed = Signal(float)
    threshold_loc_changed = Signal(float)
    threshold_changepoint_changed = Signal(float)
    segment_length_threshold_changed = Signal(float)

    def __init__(self):
        super().__init__()
        self._threshold_temp = None
        self._threshold_loc = None
        self._threshold_changepoint = None
        self._segment_length_threshold = None

    @property
    def threshold_temp(self):
        return self._threshold_temp

    @threshold_temp.setter
    def threshold_temp(self, value):
        self._threshold_temp = value
        self.threshold_temp_changed.emit(value)

    @property
    def threshold_loc(self):
        return self._threshold_loc

    @threshold_loc.setter
    def threshold_loc(self, value):
        self._threshold_loc = value
        self.threshold_loc_changed.emit(value)

    @property
    def threshold_changepoint(self):
        return self._threshold_changepoint

    @threshold_changepoint.setter
    def threshold_changepoint(self, value):
        self._threshold_changepoint = value
        self.threshold_changepoint_changed.emit(value)

    @property
    def segment_length_threshold(self):
        return self._segment_length_threshold

    @segment_length_threshold.setter
    def segment_length_threshold(self, value):
        self._segment_length_threshold = value
        self.segment_length_threshold_changed.emit(value)


# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'mainwindow.ui'
##
## Created by: Qt User Interface Compiler version 6.2.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWebEngineWidgets import QWebEngineView
from PySide6.QtWidgets import (QApplication, QGridLayout, QMainWindow, QMenu,
    QMenuBar, QSizePolicy, QStatusBar, QWidget)

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.setEnabled(True)
        MainWindow.resize(800, 600)
        self.actionOpen_Dataset = QAction(MainWindow)
        self.actionOpen_Dataset.setObjectName(u"actionOpen_Dataset")
        self.actionClose_Dataset = QAction(MainWindow)
        self.actionClose_Dataset.setObjectName(u"actionClose_Dataset")
        self.actionClose_Dataset.setEnabled(False)
        self.actionQuit = QAction(MainWindow)
        self.actionQuit.setObjectName(u"actionQuit")
        self.actionAnnotation_Editor = QAction(MainWindow)
        self.actionAnnotation_Editor.setObjectName(u"actionAnnotation_Editor")
        self.actionAbout = QAction(MainWindow)
        self.actionAbout.setObjectName(u"actionAbout")
        self.actionNew_Defect_Annotation = QAction(MainWindow)
        self.actionNew_Defect_Annotation.setObjectName(u"actionNew_Defect_Annotation")
        self.actionNew_Defect_Annotation.setEnabled(False)
        self.actionSave_Defect_Annotation = QAction(MainWindow)
        self.actionSave_Defect_Annotation.setObjectName(u"actionSave_Defect_Annotation")
        self.actionSave_Defect_Annotation.setEnabled(False)
        self.actionLoad_Defect_Annotation = QAction(MainWindow)
        self.actionLoad_Defect_Annotation.setObjectName(u"actionLoad_Defect_Annotation")
        self.actionLoad_Defect_Annotation.setEnabled(False)
        self.actionNew_Analysis = QAction(MainWindow)
        self.actionNew_Analysis.setObjectName(u"actionNew_Analysis")
        self.actionNew_Analysis.setEnabled(False)
        self.actionAnnotate_Strings = QAction(MainWindow)
        self.actionAnnotate_Strings.setObjectName(u"actionAnnotate_Strings")
        self.actionAnnotate_Strings.setEnabled(False)
        self.actionClose_String_Annotation = QAction(MainWindow)
        self.actionClose_String_Annotation.setObjectName(u"actionClose_String_Annotation")
        self.actionClose_String_Annotation.setEnabled(False)
        self.actionSave_String_Annotation = QAction(MainWindow)
        self.actionSave_String_Annotation.setObjectName(u"actionSave_String_Annotation")
        self.actionSave_String_Annotation.setEnabled(False)
        self.actionDiscard_Defect_Annotation = QAction(MainWindow)
        self.actionDiscard_Defect_Annotation.setObjectName(u"actionDiscard_Defect_Annotation")
        self.actionDiscard_Defect_Annotation.setEnabled(False)
        self.actionDiscard_String_Annotation = QAction(MainWindow)
        self.actionDiscard_String_Annotation.setObjectName(u"actionDiscard_String_Annotation")
        self.actionDiscard_String_Annotation.setEnabled(False)
        self.actionClose_Defect_Annotation = QAction(MainWindow)
        self.actionClose_Defect_Annotation.setObjectName(u"actionClose_Defect_Annotation")
        self.actionClose_Defect_Annotation.setEnabled(False)
        self.actionExport_String_Annotation = QAction(MainWindow)
        self.actionExport_String_Annotation.setObjectName(u"actionExport_String_Annotation")
        self.actionExport_String_Annotation.setEnabled(False)
        self.actionDataset_Settings = QAction(MainWindow)
        self.actionDataset_Settings.setObjectName(u"actionDataset_Settings")
        self.actionDataset_Settings.setEnabled(False)
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.gridLayout = QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.widget = QWebEngineView(self.centralwidget)
        self.widget.setObjectName(u"widget")

        self.gridLayout.addWidget(self.widget, 0, 0, 1, 1)

        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(MainWindow)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setGeometry(QRect(0, 0, 800, 22))
        self.menuFile = QMenu(self.menubar)
        self.menuFile.setObjectName(u"menuFile")
        self.menuView = QMenu(self.menubar)
        self.menuView.setObjectName(u"menuView")
        self.menuHelp = QMenu(self.menubar)
        self.menuHelp.setObjectName(u"menuHelp")
        self.menuAnalysis = QMenu(self.menubar)
        self.menuAnalysis.setObjectName(u"menuAnalysis")
        self.menuAnnotation = QMenu(self.menubar)
        self.menuAnnotation.setObjectName(u"menuAnnotation")
        MainWindow.setMenuBar(self.menubar)
        self.statusBar = QStatusBar(MainWindow)
        self.statusBar.setObjectName(u"statusBar")
        MainWindow.setStatusBar(self.statusBar)

        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuAnnotation.menuAction())
        self.menubar.addAction(self.menuAnalysis.menuAction())
        self.menubar.addAction(self.menuView.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())
        self.menuFile.addAction(self.actionOpen_Dataset)
        self.menuFile.addAction(self.actionDataset_Settings)
        self.menuFile.addAction(self.actionClose_Dataset)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionQuit)
        self.menuHelp.addAction(self.actionAbout)
        self.menuAnalysis.addAction(self.actionNew_Analysis)
        self.menuAnnotation.addAction(self.actionNew_Defect_Annotation)
        self.menuAnnotation.addAction(self.actionLoad_Defect_Annotation)
        self.menuAnnotation.addAction(self.actionSave_Defect_Annotation)
        self.menuAnnotation.addAction(self.actionClose_Defect_Annotation)
        self.menuAnnotation.addSeparator()
        self.menuAnnotation.addAction(self.actionAnnotate_Strings)
        self.menuAnnotation.addAction(self.actionExport_String_Annotation)
        self.menuAnnotation.addAction(self.actionClose_String_Annotation)

        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"PV Hawk Viewer", None))
        self.actionOpen_Dataset.setText(QCoreApplication.translate("MainWindow", u"Open Dataset...", None))
        self.actionClose_Dataset.setText(QCoreApplication.translate("MainWindow", u"Close Dataset", None))
        self.actionQuit.setText(QCoreApplication.translate("MainWindow", u"Quit", None))
        self.actionAnnotation_Editor.setText(QCoreApplication.translate("MainWindow", u"Annotation Editor", None))
        self.actionAbout.setText(QCoreApplication.translate("MainWindow", u"About", None))
        self.actionNew_Defect_Annotation.setText(QCoreApplication.translate("MainWindow", u"New Defect Annotation...", None))
        self.actionSave_Defect_Annotation.setText(QCoreApplication.translate("MainWindow", u"Save Defect Annotation", None))
        self.actionLoad_Defect_Annotation.setText(QCoreApplication.translate("MainWindow", u"Load Defect Annotation...", None))
        self.actionNew_Analysis.setText(QCoreApplication.translate("MainWindow", u"New Analysis...", None))
        self.actionAnnotate_Strings.setText(QCoreApplication.translate("MainWindow", u"Annotate Strings...", None))
        self.actionClose_String_Annotation.setText(QCoreApplication.translate("MainWindow", u"Close String Annotation", None))
        self.actionSave_String_Annotation.setText(QCoreApplication.translate("MainWindow", u"Save String Annotation", None))
        self.actionDiscard_Defect_Annotation.setText(QCoreApplication.translate("MainWindow", u"Discard Defect Annotation", None))
        self.actionDiscard_String_Annotation.setText(QCoreApplication.translate("MainWindow", u"Discard String Annotation", None))
        self.actionClose_Defect_Annotation.setText(QCoreApplication.translate("MainWindow", u"Close Defect Annotation", None))
        self.actionExport_String_Annotation.setText(QCoreApplication.translate("MainWindow", u"Export String Annotation...", None))
        self.actionDataset_Settings.setText(QCoreApplication.translate("MainWindow", u"Dataset Settings", None))
        self.menuFile.setTitle(QCoreApplication.translate("MainWindow", u"File", None))
        self.menuView.setTitle(QCoreApplication.translate("MainWindow", u"View", None))
        self.menuHelp.setTitle(QCoreApplication.translate("MainWindow", u"Help", None))
        self.menuAnalysis.setTitle(QCoreApplication.translate("MainWindow", u"Analysis", None))
        self.menuAnnotation.setTitle(QCoreApplication.translate("MainWindow", u"Annotation", None))
    # retranslateUi




from PySide6.QtWidgets import QWidget
from PySide6.QtCore import Qt, Slot

from ..ui.ui_analysis_details import Ui_AnalysisDetails


class AnalysisDetailsView(QWidget):
    def __init__(self, model, controller, parent=None):
        super().__init__(parent)
        self.setWindowFlags(Qt.Window)  # open in new window
        self.model = model
        self.controller = controller
        self.ui = Ui_AnalysisDetails()
        self.ui.setupUi(self)
        self.parent = parent
        # connect signals and slots
        self.ui.pushButtonClose.clicked.connect(self.close)
        self.model.meta_changed.connect(self.update)
        self.update()

    def to_display(self, text):
        """Converts a string of the form 'my_setting' to 'My Setting'."""
        return " ".join([str.capitalize(s) for s in str.split(text, "_")])

    def truncate_string(self, value, max_len=100):
        value = str(value)
        if len(value) > max_len:
            value = value[:max_len] + "..."
        return value

    @Slot()
    def update(self):  
        meta = self.model.meta
        if meta is None:
            self.ui.lineEditAnalysisType.setText("")
            self.ui.lineEditTimestamp.setText("")
            self.ui.lineEditDatasetPath.setText("")
            self.ui.listWidgetHyperparameters.clear()
            return

        self.ui.lineEditAnalysisType.setText(self.to_display(meta["type"]))
        self.ui.lineEditTimestamp.setText(meta["timestamp"])
        self.ui.lineEditDatasetPath.setText(meta["dataset_dir"])
        hyperparameters = [
            "{}: {}".format(self.to_display(name), self.truncate_string(value)) 
            for name, value 
            in meta["hyperparameters"].items()
        ]
        self.ui.listWidgetHyperparameters.clear()
        self.ui.listWidgetHyperparameters.addItems(hyperparameters)


# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'string_editor.ui'
##
## Created by: Qt User Interface Compiler version 6.2.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QGridLayout, QGroupBox, QHBoxLayout,
    QLabel, QLineEdit, QPushButton, QSizePolicy,
    QSpacerItem, QWidget)

class Ui_StringEditor(object):
    def setupUi(self, StringEditor):
        if not StringEditor.objectName():
            StringEditor.setObjectName(u"StringEditor")
        StringEditor.resize(446, 333)
        self.gridLayout_2 = QGridLayout(StringEditor)
        self.gridLayout_2.setSpacing(0)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(-1, -1, -1, 6)
        self.pushButtonNewString = QPushButton(StringEditor)
        self.pushButtonNewString.setObjectName(u"pushButtonNewString")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushButtonNewString.sizePolicy().hasHeightForWidth())
        self.pushButtonNewString.setSizePolicy(sizePolicy)

        self.horizontalLayout.addWidget(self.pushButtonNewString)

        self.pushButtonDeleteString = QPushButton(StringEditor)
        self.pushButtonDeleteString.setObjectName(u"pushButtonDeleteString")
        sizePolicy.setHeightForWidth(self.pushButtonDeleteString.sizePolicy().hasHeightForWidth())
        self.pushButtonDeleteString.setSizePolicy(sizePolicy)

        self.horizontalLayout.addWidget(self.pushButtonDeleteString)


        self.gridLayout_2.addLayout(self.horizontalLayout, 0, 0, 1, 1)

        self.verticalSpacer = QSpacerItem(20, 0, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.gridLayout_2.addItem(self.verticalSpacer, 2, 0, 1, 1)

        self.horizontalSpacer = QSpacerItem(0, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.gridLayout_2.addItem(self.horizontalSpacer, 1, 1, 1, 1)

        self.groupBox = QGroupBox(StringEditor)
        self.groupBox.setObjectName(u"groupBox")
        sizePolicy1 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.groupBox.sizePolicy().hasHeightForWidth())
        self.groupBox.setSizePolicy(sizePolicy1)
        self.gridLayout = QGridLayout(self.groupBox)
        self.gridLayout.setObjectName(u"gridLayout")
        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.pushButtonStartDrawing = QPushButton(self.groupBox)
        self.pushButtonStartDrawing.setObjectName(u"pushButtonStartDrawing")
        sizePolicy2 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.pushButtonStartDrawing.sizePolicy().hasHeightForWidth())
        self.pushButtonStartDrawing.setSizePolicy(sizePolicy2)

        self.horizontalLayout_3.addWidget(self.pushButtonStartDrawing)

        self.pushButtonPauseDrawing = QPushButton(self.groupBox)
        self.pushButtonPauseDrawing.setObjectName(u"pushButtonPauseDrawing")
        sizePolicy2.setHeightForWidth(self.pushButtonPauseDrawing.sizePolicy().hasHeightForWidth())
        self.pushButtonPauseDrawing.setSizePolicy(sizePolicy2)

        self.horizontalLayout_3.addWidget(self.pushButtonPauseDrawing)

        self.pushButtonEndDrawing = QPushButton(self.groupBox)
        self.pushButtonEndDrawing.setObjectName(u"pushButtonEndDrawing")
        sizePolicy2.setHeightForWidth(self.pushButtonEndDrawing.sizePolicy().hasHeightForWidth())
        self.pushButtonEndDrawing.setSizePolicy(sizePolicy2)

        self.horizontalLayout_3.addWidget(self.pushButtonEndDrawing)


        self.gridLayout.addLayout(self.horizontalLayout_3, 0, 1, 1, 1)

        self.lineEditStringID = QLineEdit(self.groupBox)
        self.lineEditStringID.setObjectName(u"lineEditStringID")
        sizePolicy.setHeightForWidth(self.lineEditStringID.sizePolicy().hasHeightForWidth())
        self.lineEditStringID.setSizePolicy(sizePolicy)

        self.gridLayout.addWidget(self.lineEditStringID, 4, 1, 1, 1)

        self.label_3 = QLabel(self.groupBox)
        self.label_3.setObjectName(u"label_3")
        sizePolicy1.setHeightForWidth(self.label_3.sizePolicy().hasHeightForWidth())
        self.label_3.setSizePolicy(sizePolicy1)

        self.gridLayout.addWidget(self.label_3, 4, 0, 1, 1)

        self.lineEditTrackerID = QLineEdit(self.groupBox)
        self.lineEditTrackerID.setObjectName(u"lineEditTrackerID")
        sizePolicy.setHeightForWidth(self.lineEditTrackerID.sizePolicy().hasHeightForWidth())
        self.lineEditTrackerID.setSizePolicy(sizePolicy)
        self.lineEditTrackerID.setMaximumSize(QSize(16777215, 16777215))

        self.gridLayout.addWidget(self.lineEditTrackerID, 1, 1, 1, 1)

        self.lineEditArrayID = QLineEdit(self.groupBox)
        self.lineEditArrayID.setObjectName(u"lineEditArrayID")
        sizePolicy.setHeightForWidth(self.lineEditArrayID.sizePolicy().hasHeightForWidth())
        self.lineEditArrayID.setSizePolicy(sizePolicy)

        self.gridLayout.addWidget(self.lineEditArrayID, 2, 1, 1, 1)

        self.horizontalLayout_5 = QHBoxLayout()
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.pushButtonCancelString = QPushButton(self.groupBox)
        self.pushButtonCancelString.setObjectName(u"pushButtonCancelString")
        sizePolicy.setHeightForWidth(self.pushButtonCancelString.sizePolicy().hasHeightForWidth())
        self.pushButtonCancelString.setSizePolicy(sizePolicy)

        self.horizontalLayout_5.addWidget(self.pushButtonCancelString)

        self.pushButtonConfirmString = QPushButton(self.groupBox)
        self.pushButtonConfirmString.setObjectName(u"pushButtonConfirmString")
        sizePolicy.setHeightForWidth(self.pushButtonConfirmString.sizePolicy().hasHeightForWidth())
        self.pushButtonConfirmString.setSizePolicy(sizePolicy)

        self.horizontalLayout_5.addWidget(self.pushButtonConfirmString)


        self.gridLayout.addLayout(self.horizontalLayout_5, 5, 1, 1, 1)

        self.lineEditInverterID = QLineEdit(self.groupBox)
        self.lineEditInverterID.setObjectName(u"lineEditInverterID")
        sizePolicy.setHeightForWidth(self.lineEditInverterID.sizePolicy().hasHeightForWidth())
        self.lineEditInverterID.setSizePolicy(sizePolicy)

        self.gridLayout.addWidget(self.lineEditInverterID, 3, 1, 1, 1)

        self.label_2 = QLabel(self.groupBox)
        self.label_2.setObjectName(u"label_2")
        sizePolicy1.setHeightForWidth(self.label_2.sizePolicy().hasHeightForWidth())
        self.label_2.setSizePolicy(sizePolicy1)

        self.gridLayout.addWidget(self.label_2, 2, 0, 1, 1)

        self.label = QLabel(self.groupBox)
        self.label.setObjectName(u"label")
        sizePolicy1.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy1)

        self.gridLayout.addWidget(self.label, 1, 0, 1, 1)

        self.label_5 = QLabel(self.groupBox)
        self.label_5.setObjectName(u"label_5")

        self.gridLayout.addWidget(self.label_5, 0, 0, 1, 1)

        self.label_4 = QLabel(self.groupBox)
        self.label_4.setObjectName(u"label_4")
        sizePolicy1.setHeightForWidth(self.label_4.sizePolicy().hasHeightForWidth())
        self.label_4.setSizePolicy(sizePolicy1)

        self.gridLayout.addWidget(self.label_4, 3, 0, 1, 1)


        self.gridLayout_2.addWidget(self.groupBox, 1, 0, 1, 1)


        self.retranslateUi(StringEditor)

        QMetaObject.connectSlotsByName(StringEditor)
    # setupUi

    def retranslateUi(self, StringEditor):
        StringEditor.setWindowTitle(QCoreApplication.translate("StringEditor", u"Form", None))
        self.pushButtonNewString.setText(QCoreApplication.translate("StringEditor", u"New String...", None))
        self.pushButtonDeleteString.setText(QCoreApplication.translate("StringEditor", u"Delete String", None))
        self.groupBox.setTitle(QCoreApplication.translate("StringEditor", u"New String", None))
        self.pushButtonStartDrawing.setText(QCoreApplication.translate("StringEditor", u"Start", None))
        self.pushButtonPauseDrawing.setText(QCoreApplication.translate("StringEditor", u"Pause", None))
        self.pushButtonEndDrawing.setText(QCoreApplication.translate("StringEditor", u"End", None))
        self.label_3.setText(QCoreApplication.translate("StringEditor", u"String ID", None))
        self.pushButtonCancelString.setText(QCoreApplication.translate("StringEditor", u"Cancel String", None))
        self.pushButtonConfirmString.setText(QCoreApplication.translate("StringEditor", u"Confirm String", None))
        self.label_2.setText(QCoreApplication.translate("StringEditor", u"Array ID", None))
        self.label.setText(QCoreApplication.translate("StringEditor", u"Tracker ID", None))
        self.label_5.setText(QCoreApplication.translate("StringEditor", u"Draw String", None))
        self.label_4.setText(QCoreApplication.translate("StringEditor", u"Inverter ID", None))
    # retranslateUi




import os
import numpy as np


def get_immediate_subdirectories(a_dir):
    """Returns the immediate subdirectories of the provided directory."""
    return [name for name in os.listdir(a_dir) if os.path.isdir(os.path.join(a_dir, name))]


def to_celsius(image, gain, offset):
    """Convert raw intensity values of radiometric image to Celsius scale."""
    return image*gain + offset


def normalize(image, vmin, vmax):
    image[image < vmin] = vmin
    image[image > vmax] = vmax
    image = (image - vmin) / (vmax - vmin)
    image = (255*image).astype(np.uint8)
    return image
